# -*- encoding: utf-8 -*-
##############################################################################
#
#    Copyright (c) 2011 Ting & Avanzosc (http://www.openerpsite.com) All Rights Reserved.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

from osv import osv, fields
import tools
import math
from tools.translate import _

class bom_fixed_cost(osv.osv):
    _name = 'bom.fixed.cost'
    _description = 'Bom fixed costs'
    _columns = {
                'name':fields.char('Description', size = 256),
                'amount':fields.float('Ammount'),
                'estimated':fields.float('Estimated amount'),
                'bom_id':fields.many2one('mrp.bom','Bom')
                }

bom_fixed_cost()

class mrp_bom_operations_product(osv.osv):
    _name = 'mrp.bom.operations.product'
    _description = 'materials used to produce'
    
    def _calculate_time(self, cr, uid, ids, field_name, arg, context=None):
        res = {}
        operation = self.pool.get('mrp.bom.operations.product').browse(cr, uid, ids)[0]
        if operation:
            if operation.semi_prod_qty:
                cycle = (operation.semi_prod_qty/operation.capacity_per_cycle)
            else:
                cycle = (operation.bom_id.product_qty/operation.capacity_per_cycle)
            res[operation.id] = {
                    'cycle': cycle,
                    'hour': float(((operation.time_cycle * cycle) + operation.time_start + operation.time_stop) / operation.time_efficiency)
            }
        return res
    
    def _compute_qty(self, cr, uid, ids, field_name, arg, context=None):
        id = ids[0]
        bom_objs = self.pool.get('mrp.bom')
        operation = self.pool.get('mrp.bom.operations.product').browse(cr, uid, id)
        bom = bom_objs.browse(cr, uid, bom_objs.search(cr, uid, [('mrp_bom_operations_product_ids', 'in', id)]))[0]
        if operation.type == 'variable' and operation.semi_prod_qty > 0:
            return {id: 1 / float(bom.product_qty / operation.semi_prod_qty)}
        else:
            return {id: float(operation.semi_prod_qty)}
    
    def create(self, cr, uid, data, context=None):
        if 'name' not in data:
            op = self.pool.get('mrp.operation').browse(cr, uid, data['operation_id'])
            wc = self.pool.get('mrp.workcenter').browse(cr, uid, data['workcenter_id'])
            data['name'] = '%s -> %s' % (op.name, wc.name)
        return super(mrp_bom_operations_product, self).create(cr, uid, data, context)
    
    _columns = { 
        'name': fields.char('Name', size = 64, required= True),    
        'sequence': fields.integer('Sequence', required= True),
        'descript': fields.text('Description', size= 64),    
        'product_id': fields.many2one('product.product', 'Product'),
        'semi_prod_qty': fields.float('Product Qty'),
        'semi_prod_uom': fields.many2one('product.uom', 'Product UOM'),
        'semi_prod_qty_fact': fields.function(_compute_qty, string='Quantity Factor', method=True, type='float',
            store={'mrp.bom.operations.product': (lambda self, cr, uid, ids, c={}: ids, None, 20)}),
        'operation_id': fields.many2one('mrp.operation', 'Operation', required=True),
        'workcenter_id': fields.many2one('mrp.workcenter', 'Work Center', required=True),
        'location_id': fields.many2one('stock.location', 'Operation Location', required= True, domain=[('usage','in',('internal','supplier'))]),
        'capacity_per_cycle': fields.float('Capacity per Cycle'),
        'time_efficiency' : fields.float('Efficiency factor', required=True),
        'time_cycle': fields.float('Time for 1 cycle (hour)', help="Time in hours for doing one cycle of the material."),
        'time_start': fields.float('Time before prod.', help="Time in hours for the setup for the material."),
        'time_stop': fields.float('Time after prod.', help="Time in hours for the cleaning."),
        'costs_hour': fields.float('Cost per hour', help="Specify Cost of Workcenter per hour when using this material."),
        'costs_cycle': fields.float('Cost per cycle', help="Specify Cost of Workcenter per cycle when using this material."),
        'cycle': fields.function(_calculate_time, string='Nº Cycles', method=True, type='float',
            store={'mrp.bom.operations.product': (lambda self, cr, uid, ids, c={}: ids, None, 20)}, multi="All"),
        'hour': fields.function(_calculate_time, string='Nº Hour', method=True, type='float',
            store={'mrp.bom.operations.product': (lambda self, cr, uid, ids, c={}: ids, None, 20)}, multi="All"),
        'final_products_ids': fields.one2many('mrp.bom.final.product', 'mrp_bom_operations_product_id', 'Final products'),        
        'components_used_ids': fields.one2many('mrp.bom', 'mrp_bom_operations_product_id', 'Components used'), 
        'bom_id': fields.many2one('mrp.bom', 'Parent BOM', required=True),
        'type': fields.selection([('fixed','Fixed'),('variable','Variable')],'Type'),     
        
    }
    
    _defaults = {
        'sequence': lambda *a: 1,
        'time_efficiency': lambda *a: 1,
        'cycle': lambda *a: 1.0,
        'hour': lambda *a: 0.0,
        'type': lambda *a: 'variable',
    }
    
    def default_get(self, cr, uid, fields, context=None):
        if context is None:
            context = {}
        res = super(mrp_bom_operations_product, self).default_get(cr, uid, fields, context)
        if 'routing_id' in context.keys():
            id = self.pool.get('mrp.bom').search(cr, uid, [('name', '=', context['name']), ('routing_id', '=', context['routing_id'])])[0]
            res.update({'bom_id': id})
        return res
    
    def wc_onchange(self, cr, uid, ids, workcenter_id, operation_id):
        value = {}
        if workcenter_id:
            wc = self.pool.get('mrp.workcenter').browse(cr, uid, workcenter_id)
            op = self.pool.get('mrp.operation').browse(cr, uid, operation_id)
            value['location_id'] = wc.location_id.id
            value['capacity_per_cycle'] = wc.capacity_per_cycle
            value['time_efficiency'] = wc.time_efficiency
            value['time_cycle'] = wc.time_cycle
            value['time_start'] = wc.time_start
            value['time_stop'] = wc.time_stop
            value['costs_hour'] = wc.costs_hour
            value['costs_cycle'] = wc.costs_cycle
            value['name'] = '%s -> %s' % (op.name, wc.name)
        return {'value': value}
    
    def prod_onchange(self, cr, uid, ids, product_id):
        value = {}
        warning = {}
        if product_id:
            prod = self.pool.get('product.product').browse(cr, uid, product_id)
            if prod.default_location.usage != 'production':
                value.update({
                    'product_id': self.browse(cr, uid, ids[0]).product_id.id,
                })
                warning = {
                    'title': _("Wrong Product"),
                    'message': _("This is not a valid product. Please, select a product with 'production' location type."),
                }
            else:
                value.update({
                    'semi_prod_qty_fact': 1,
                    'semi_prod_uom': prod.product_tmpl_id.uom_id.id or False,
                })
        return {'value': value, 'warning': warning}
    
    def cycle_onchange(self, cr, uid, ids, time_cycle, costs_hour):
        value = {}
        costh = costs_hour * time_cycle
        value['costs_cycle'] = costh
        return {'value':value}
        
    def button_dummy(self, cr, uid, ids, *args):
        self.pool.get('mrp.bom.operations.product').write(cr, uid, ids, {})
        return True
mrp_bom_operations_product()

class mrp_bom_final_product(osv.osv):
    _name = 'mrp.bom.final.product'
    _description = 'Bill of Material'
    
    _columns = {
        'product_id': fields.many2one('product.product', 'Product', required=True),
        'product_qty': fields.float('Product Qty', required=True),
        'product_uom': fields.many2one('product.uom', 'Product UOM', required=True, help="UoM (Unit of Measure) is the unit of measurement for the inventory control"),
        'mrp_bom_operations_product_id': fields.many2one('mrp.bom.operations.product','Produced On'),
        'location_id': fields.related('product_id', 'default_location', type='many2one', relation='stock.location', string='Default Location', required=True),
        'bom_id': fields.many2one('mrp.bom', 'Final Product'),
    }
    
    _defaults = {  
        'product_qty': lambda *a: 1,
        'product_uom': lambda *a: 1,
    }
    
    def default_get(self, cr, uid, fields, context=None):
        if context is None:
            context = {}
        res = super(mrp_bom_final_product, self).default_get(cr, uid, fields, context)
        if 'routing_id' in context.keys():
            id = self.pool.get('mrp.bom').search(cr, uid, [('name', '=', context['name']), ('routing_id', '=', context['routing_id'])])
            res.update({'bom_id': id})
        if 'bom_id' in context.keys():
            res.update({'bom_id': context['bom_id']})
        return res
    
    def prod_onchange(self, cr, uid, ids, product_id):
        value = {}
        if product_id:
            prod = self.pool.get('product.product').browse(cr, uid, product_id)
            value['location_id'] = prod.default_location.id
            value['product_uom'] = prod.product_tmpl_id.uom_id.id or False
        return {'value': value}
    
mrp_bom_final_product()

def rounding(f, r):
    if not r:
        return f
    return round(f / r) * r

class mrp_bom(osv.osv):
    _description = 'Bill of Material'
    _inherit = 'mrp.bom'
    
#    def write(self, cr, uid, ids, value, context=None):
#        id = ids[0]
#        for line in value['bom_lines']:
#            print id, line
#            if line[2] != False:
#                line[2]['bom_id'] = id
#            if line[0] == 0:
#                find_bom_id = self.pool.get('mrp.bom')._bom_find(cr, uid, line[2]['product_id'], line[2]['product_uom'])
#                print find_bom_id
#                if find_bom_id:
#                    return super(mrp_bom, self).write(cr, uid, find_bom_id, line[2], context)
#                else:
#                    return self.pool.get('mrp.bom').create(cr, uid, line[2])
#        return True
    
    def copy(self, cr, uid, id, default=None, context=None):
        if default is None:
            default = {}       
        
        bom_super = super(mrp_bom, self).copy(cr, uid, id, default, context)
        
        original_bom = self.browse(cr, uid, id)
        copy_bom_id = self.search(cr, uid, [('name','=',(original_bom.name+' Copy'))])[0]
        copy_bom  = self.browse(cr, uid, copy_bom_id)
        for bom_aux in copy_bom.bom_lines:
            aux_bom_op = self.pool.get('mrp.bom.operations.product').browse(cr, uid, bom_aux.mrp_bom_operations_product_id)[0]
            for aux_op in bom_aux.mrp_bom_operations_product_ids:
                if aux_op.name == aux_bom_op.name:
                    bom_aux.write({'mrp_bom_operations_product_id':aux_op.id})
        return bom_super	
    
    
    _columns = {
        'mrp_bom_operations_product_ids': fields.one2many('mrp.bom.operations.product', 'bom_id', 'Operations'),
        'mrp_bom_operations_product_id': fields.many2one('mrp.bom.operations.product','Consumed On'),
        'final_product_line_ids': fields.one2many('mrp.bom.final.product', 'bom_id', 'Final Product Lines'),
        'routing_id_name': fields.char('Routing',size=64),   
        'fixed_costs':fields.one2many('bom.fixed.cost','bom_id','Fixed costs')
    }
    
    def default_get(self, cr, uid, fields, context=None):
        if context is None:
            context = {}
        res = super(mrp_bom, self).default_get(cr, uid, fields, context)
        if 'routing_id' in context.keys():
            id = self.search(cr, uid, [('name', '=', context['name']), ('routing_id', '=', context['routing_id'])])
            res.update({'bom_id': id})
        if 'bom_id' in context.keys():
            res.update({'bom_id': context['bom_id']})
        return res
    
    def prod_onchange(self, cr, uid, ids, product_id):
        value = {}
        if product_id:
            prod = self.pool.get('product.product').browse(cr, uid, product_id)
#            value['location_id'] = prod.default_location.id
            value['name'] = prod.name
            value['product_uom'] = prod.product_tmpl_id.uom_id.id or False
        return {'value': value}
    
    def onchange_routingField(self, cr, uid, ids, routing_id):
        operations_product_obj = self.pool.get('mrp.bom.operations.product')
        if not self.browse(cr, uid, ids):
          raise osv.except_osv(_('User error'), _('Save the record first !'))
        mrp_bom_id = self.browse(cr, uid, ids)[0]
        workcenter_lines_obj = self.pool.get('mrp.production.workcenter.line')
        workcenter_line_ids = workcenter_lines_obj.search(cr, uid, [('id', '=', routing_id)])
        workcenter_line_objs = []
        
        if len(workcenter_line_ids):
            workcenter_line_objs = workcenter_lines_obj.browse(cr, uid, workcenter_line_ids)
        
        for operation_product_id in mrp_bom_id.mrp_bom_operations_product_ids:
            operations_product_obj.unlink(cr, uid, [operation_product_id.id])
        
        values = {
                'sequence': 0,
                'descript': '',
                'operation_id': None,
                'location_id': None,
                'workcenter_id': None,
                'bom_id': mrp_bom_id.id,
                'capacity_per_cycle': 0,
                'time_efficiency': 1,
                'time_cycle': 0,
                'time_start': 0,
                'time_stop': 0,
                'costs_hour': 0,
                'costs_cycle': 0,
        }
        
        for workcenter_line in workcenter_line_objs:
            values['sequence'] = workcenter_line.sequence
            values['descript'] = workcenter_line.descript
            values['operation_id'] = workcenter_line.operation_id.id
            values['location_id'] = workcenter_line.workcenter_id.location_id.id
            values['workcenter_id'] = workcenter_line.workcenter_id.id
            values['capacity_per_cycle'] = workcenter_line.workcenter_id.capacity_per_cycle
            values['time_efficiency'] = workcenter_line.workcenter_id.time_efficiency
            values['time_cycle'] = workcenter_line.workcenter_id.time_cycle
            values['time_start'] = workcenter_line.workcenter_id.time_start
            values['time_stop'] = workcenter_line.workcenter_id.time_stop
            values['costs_hour'] = workcenter_line.workcenter_id.costs_hour
            values['costs_cycle'] = workcenter_line.workcenter_id.costs_cycle
            values['name'] = '%s -> %s' % (workcenter_line.operation_id.name, workcenter_line.workcenter_id.name)
            res = operations_product_obj.create(cr, uid, values)
    
        return {}
    
    def _bom_explode(self, cr, uid, bom, factor, properties=[], addthis=False, level=0):
        """ Finds Products and Workcenters for related BoM for manufacturing order.
        @param bom: BoM of particular product.
        @param factor: Factor of product UoM.
        @param properties: A List of properties Ids.
        @param addthis: If BoM found then True else False.
        @param level: Depth level to find BoM lines starts from 10.
        @return: result: List of dictionaries containing product details.
                 result2: List of dictionaries containing workcenter details.
        """
        factor = factor / (bom.product_efficiency or 1.0)
        factor = rounding(factor, bom.product_rounding)
        if factor < bom.product_rounding:
            factor = bom.product_rounding
        result = []
        phantom = False
        if bom.type == 'phantom' and not bom.bom_lines:
            newbom = self._bom_find(cr, uid, bom.product_id.id, bom.product_uom.id, properties)
            if newbom:
                res = self._bom_explode(cr, uid, self.browse(cr, uid, [newbom])[0], factor*bom.product_qty, properties, addthis=True, level=level+10)
                result = result + res[0]
                phantom = True
            else:
                phantom = False
        if not phantom:            
            if bom.routing_id:
                for op_use in bom.mrp_bom_operations_product_ids:
                    products = []
                    finals = []
                    for product in op_use.components_used_ids:
                        if product.product_uom.name == 'PCE': 
                            qty = math.ceil(product.product_qty )
                        else:
                            qty = product.product_qty
                        if op_use.type == 'variable':
                            qty = qty * factor
                                                         
                        products.append({
                            'name': product.product_id.name,
                            'product_id': product.product_id.id,
                            'product_qty': qty,
                            'product_uom': product.product_uom.id,
                            'product_uos_qty': product.product_uos and bom.product_uos_qty * factor or False,
                            'product_uos': product.product_uos and bom.product_uos.id or False,
                        })
                    for final in op_use.final_products_ids:                            
                        if final.product_uom.name == 'PCE':
                            qty = math.ceil(final.product_qty )
                        else:                       
                            qty = final.product_qty                       
                        if op_use.type == 'variable':
                            qty = qty * factor
                            
                        finals.append({
                            'name': final.product_id.name,
                            'product_id': final.product_id.id,
                            'product_qty': qty,
                            'product_uom': final.product_uom.id,
                            'location_id': final.location_id.id,
                        })
                    result.append({
                        'name': tools.ustr(op_use.name),
                        'workcenter_id': op_use.workcenter_id.id,
                        'operation_id': op_use.operation_id.id,
                        'location_id': op_use.location_id.id,
                        'type': op_use.type,
                        'product_id': op_use.product_id.id,
                        'semi_prod_uom': op_use.semi_prod_uom.id,
                        'semi_prod_qty_fact': op_use.semi_prod_qty_fact or 0,
                        'sequence': level+(op_use.sequence or 0),
                        'capacity_per_cycle': op_use.capacity_per_cycle,
                        'time_efficiency': op_use.time_efficiency,
                        'time_cycle': op_use.time_cycle,
                        'time_start': op_use.time_start,
                        'time_stop': op_use.time_stop,
                        'costs_hour': op_use.costs_hour,
                        'costs_cycle': op_use.costs_cycle,
                        'hour': op_use.hour,
                        'cycle': op_use.cycle,
                        'products': products,
                        'finals': finals,
                    })
            for bom2 in bom.bom_lines:
                res = self._bom_explode(cr, uid, bom2, factor, properties, addthis=True, level=level+10)
                if res != []:
                    result = result + res[0]
        return result
    
mrp_bom()
