# -*- encoding: utf-8 -*-
##############################################################################
#
#    Copyright (c) 2011 Ting & Avanzosc (http://www.openerpsite.com) All Rights Reserved.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see http://www.gnu.org/licenses/.
#
##############################################################################

from datetime import datetime
from dateutil.relativedelta import relativedelta

from osv import osv, fields
from tools.translate import _
import netsvc


class mrp_fixed_cost(osv.osv):
    _name = 'mrp.fixed.cost'
    _description = 'MRP fixed costs'
    _columns = {
                'name':fields.char('Description', size = 256),
                'amount':fields.float('Ammount'),
                'estimated':fields.float('Estimated amount'),
                'production_id':fields.many2one('mrp.production','Bom'),
                'created':fields.boolean ('created')
                }
    _defaults = {
                 'created': False
                 }

mrp_fixed_cost()

class mrp_production_operation(osv.osv):
    _name = 'mrp.production.operation'
    _description = 'Production Operation'
    
    def _calculate_time(self, cr, uid, ids, field_name, arg, context=None):
        res = {}
        operation = self.pool.get('mrp.production.operation').browse(cr, uid, ids)[0]
#        production = self.pool.get('mrp.production').browse(cr, uid, operation.production_id.id)
        if operation:
            if operation.semi_prod_qty:
                cycle = (operation.semi_prod_qty/operation.capacity_per_cycle)
            else:
                cycle = (operation.production_id.product_qty/operation.capacity_per_cycle)
            res[operation.id] = {
                    'cycle': cycle,
                    'hour': float(((operation.time_cycle * cycle) + operation.time_start + operation.time_stop) / operation.time_efficiency)
            }
        return res
    
    def _compute_qty(self, cr, uid, ids, field_name, arg, context=None):
        id = ids[0]
        production_objs = self.pool.get('mrp.production')
        operation = self.pool.get('mrp.production.operation').browse(cr, uid, id)
        production = production_objs.browse(cr, uid, production_objs.search(cr, uid, [('production_operation_ids', 'in', id)]))[0]
        if operation.type == 'variable':
            return {id: float(production.product_qty * operation.semi_prod_qty_fact)}
        else:
            return {id: float(operation.semi_prod_qty_fact)}
    
    _columns = {
            'name': fields.char('Name', size = 64, required= True),
            'sequence': fields.integer('Sequence', size = 64, required= True),
            'descript': fields.text('Description', size= 64),    
            'product_id': fields.many2one('product.product', 'Semi-Finished Product'),
            'semi_prod_qty': fields.function(_compute_qty, string='Product Qty', method=True, type='float',
                store={'mrp.production.operation': (lambda self, cr, uid, ids, c={}: ids, None, 20)}),
            'semi_prod_uom': fields.many2one('product.uom', 'Product UOM'),
            'semi_prod_qty_fact': fields.float('Quantity Factor', size = 64),
            'operation_id': fields.many2one('mrp.operation', 'Operation', required=True),
            'location_id': fields.many2one('stock.location', 'Operation Location', required= True, domain=[('usage','in',('internal', 'supplier'))]),
            'workcenter_id': fields.many2one('mrp.workcenter', 'Work Center', required=True),
            'production_id': fields.many2one('mrp.production', 'Work Order', required=True),
            'capacity_per_cycle': fields.float('Capacity per Cycle'),
            'time_efficiency' : fields.float('Efficiency factor', required=True),
            'time_cycle': fields.float('Time for 1 cycle (hour)', help="Time in hours for doing one cycle of the material."),
            'time_start': fields.float('Time before prod.', help="Time in hours for the setup for the material."),
            'time_stop': fields.float('Time after prod.', help="Time in hours for the cleaning."),
            'costs_hour': fields.float('Cost per hour', help="Specify Cost of Workcenter per hour when using this material."),
            'costs_cycle': fields.float('Cost per cycle', help="Specify Cost of Workcenter per cycle when using this material."),
            'cycle': fields.function(_calculate_time, string='Nº Cycles', method=True, type='float',
                store={'mrp.production.operation': (lambda self, cr, uid, ids, c={}: ids, None, 20)}, multi="All"),
            'hour': fields.function(_calculate_time, string='Nº Hour', method=True, type='float',
                store={'mrp.production.operation': (lambda self, cr, uid, ids, c={}: ids, None, 20)}, multi="All"),
            'final_products_ids': fields.one2many('mrp.production.final.product', 'mrp_production_operation_id', 'Final products'),        
            'components_used_ids': fields.one2many('mrp.production.product.line', 'mrp_production_operation_id', 'Components used'),
            'type':fields.selection([('fixed','Fixed'),('variable','Variable')],'Type'),
      
    }
    
    _defaults = {
            'sequence': lambda *a: 1,
            'semi_prod_qty_fact': lambda *a: 1,
            'time_efficiency': lambda *a: 0,
            'cycle': lambda *a: 1.0,
            'hour': lambda *a: 0.0,
            'type': lambda *a: 'variable',
    }
    
    def default_get(self, cr, uid, fields, context=None):
        if context is None:
            context = {}
        res = super(mrp_production_operation, self).default_get(cr, uid, fields, context)
        if 'name' in context.keys():
            production_id = self.pool.get('mrp.production').search(cr, uid, [('name', '=', context['name'])])[0]
            res.update({'production_id': production_id})
        return res
    
    def wc_onchange(self, cr, uid, ids, workcenter_id, operation_id):
        value = {}
        if workcenter_id:
            wc = self.pool.get('mrp.workcenter').browse(cr, uid, workcenter_id)
            op = self.pool.get('mrp.operation').browse(cr, uid, operation_id)
            value['location_id'] = wc.location_id.id
            value['capacity_per_cycle'] = wc.capacity_per_cycle
            value['time_efficiency'] = wc.time_efficiency
            value['time_cycle'] = wc.time_cycle
            value['time_start'] = wc.time_start
            value['time_stop'] = wc.time_stop
            value['costs_hour'] = wc.costs_hour
            value['costs_cycle'] = wc.costs_cycle
            value['name'] = '%s -> %s' % (op.name, wc.name)
        return {'value': value}
    
    def prod_onchange(self, cr, uid, ids, product_id):
        value = {}
        warning = {}
        if product_id:
            prod = self.pool.get('product.product').browse(cr, uid, product_id)
            if prod.default_location.usage != 'production':
                value.update({
                    'product_id': self.browse(cr, uid, ids[0]).product_id.id,
                })
                warning = {
                    'title': _("Wrong Product"),
                    'message': _("This is not a valid product. Please, select a product with 'production' location type."),
                }
            else:
                value.update({
                    'semi_prod_qty_fact': 1,
                    'semi_prod_uom': prod.product_tmpl_id.uom_id.id or False,
                })
        return {'value': value, 'warning': warning}
    
    def cycle_onchange(self, cr, uid, ids, time_cycle, costs_hour):
        value = {}
        costh = costs_hour * time_cycle
        value['costs_cycle'] = costh
        return {'value':value}
    
    def button_dummy(self, cr, uid, ids, *args):
        self.pool.get('mrp.production.operation').write(cr, uid, ids, {})
        return True
mrp_production_operation()

class mrp_production_product_line(osv.osv):
    _description = 'Production Scheduled Product'
    _inherit = 'mrp.production.product.line'
    
    _columns = {
        'mrp_production_operation_id': fields.many2one('mrp.production.operation', 'Consumed On'),
        'location_id': fields.related('product_id', 'default_location', type='many2one', relation='stock.location', string='Default Location'),
    }
    
    _default = {
        'product_qty': lambda *a: 1,
    }
    
    def prod_onchange(self, cr, uid, ids, product_id):
        value = {}
        if product_id:
            prod = self.pool.get('product.product').browse(cr, uid, product_id)
            value['name'] = prod.name
            value['product_qty'] = 1
            value['product_uom'] = prod.product_tmpl_id.uom_id.id or False
            value['location_id'] = prod.default_location.id
        return {'value': value}
    
    def default_get(self, cr, uid, fields, context=None):
        if context is None:
            context = {}
        if 'production_id' in context.keys():
            return {'production_id': context['production_id']}
        if 'name' in context.keys():
            production_id = self.pool.get('mrp.production').search(cr, uid, [('name', '=', context['name'])])
            return {'production_id': production_id}
        return {}
mrp_production_product_line()

class mrp_production_final_product(osv.osv):
    _name = 'mrp.production.final.product'
    _description = 'Production Scheduled Final Product'
    
    _columns = {
        'name': fields.char('Name', size = 64, required= True),
        'product_id': fields.many2one('product.product', 'Product', required=True),
        'product_qty': fields.float('Product Qty', required=True),
        'product_uom': fields.many2one('product.uom', 'Product UOM', required=True, help="UoM (Unit of Measure) is the unit of measurement for the inventory control"),
        'mrp_production_operation_id': fields.many2one('mrp.production.operation','Produced On'),
        'location_id': fields.related('product_id', 'default_location', type='many2one', relation='stock.location', string='Default Location'),
        'production_id': fields.many2one('mrp.production', 'Final Product'),
    }
    
    def prod_onchange(self, cr, uid, ids, product_id):
        value = {}
        if product_id:
            prod = self.pool.get('product.product').browse(cr, uid, product_id)
            value['name'] = prod.name
            value['product_qty'] = 1
            value['product_uom'] = prod.product_tmpl_id.uom_id.id or False
            value['location_id'] = prod.default_location.id
        return {'value': value}
    
    def default_get(self, cr, uid, fields, context=None):
        if context is None:
            context = {}
        if 'production_id' in context.keys():
            return {'production_id': context['production_id']}
        if 'name' in context.keys():
            production_id = self.pool.get('mrp.production').search(cr, uid, [('name', '=', context['name'])])
            return {'production_id': production_id}
        return {}
    
mrp_production_final_product()

class mrp_production_users(osv.osv):
    _name = 'mrp.production.users'    
    
    def delete_it(self, cr, uid, ids, context=None):
              self.unlink(cr, uid,ids, context=context)
              return False
    
    _columns = {       
            'name': fields.many2one('res.users','User'),
            'op': fields.many2one('mrp.production.operation','Operation'),
            'prod': fields.many2one('mrp.production','Production'),
            'percent': fields.float('% Estimated'),
            'est_time':fields.float('Time Estimated'),
        }
    
    _default= {
        'percent': 0
     }
    
    def name_get(self, cr, uid, ids, context=None):
        if not ids:
            return []
        result = []
        for line in self.browse(cr, uid, ids, context=context):
            if line.name:
                result.append((line.id, (line.name.name or '')))
            else:
                result.append((line.id, line.name))
        return result
    
    
mrp_production_users()

class mrp_production(osv.osv):
    _inherit = "mrp.production"    
    _columns = {
            'bom_id': fields.many2one('mrp.bom', 'Bill of Material', readonly=True, states={'draft': [('readonly', False)]}, domain=[('bom_id','=',False)]),
            'routing_id': fields.many2one('mrp.routing', 'Routing', on_delete='set null', readonly=True, help="The list of operations (list of work centers) to produce the finished product. The routing is mainly used to compute work center costs during operations and to plan future loads on work centers based on production plannification."),
            'production_operation_ids': fields.one2many('mrp.production.operation', 'production_id', 'Production Operations', readonly=True, states={'draft': [('readonly', False)]}),
            'product_lines': fields.one2many('mrp.production.product.line', 'production_id', 'Scheduled goods', readonly=True, states={'draft': [('readonly', False)]}),
            'production_final_product_ids': fields.one2many('mrp.production.final.product', 'production_id', 'Production Final Products', readonly=True, states={'draft': [('readonly', False)]}),
            'user_ops': fields.one2many('mrp.production.users', 'prod', 'Users asignments', readonly=True, states={'draft': [('readonly', False)]}),
            'product_def_loc': fields.many2one('stock.location', 'Default Location', required=True, readonly=True, states={'draft': [('readonly', False)]}, help='Default real location of the product'),
            'location_src_id': fields.many2one('stock.location', 'Raw Materials Location', required=False,),
            'location_dest_id': fields.many2one('stock.location', 'Finished Products Location', required=False,),
            'fixed_costs':fields.one2many('mrp.fixed.cost','production_id','Fixed costs'),
            'company_id': fields.many2one('res.company', 'Company'),
    }
    
    _defaults = {  
         'company_id': lambda self, cr, uid, context: self.pool.get('res.company')._company_default_get(cr, uid, 'mrp.routing', context=context)
     }
    
    def write(self, cr, uid, ids, vals, context=None):
        res = {}
        if 'bom_id' in vals:
            for order in self.browse(cr, uid, ids):            
                if vals['bom_id']:
                    bom_id = self.pool.get('mrp.bom').browse(cr, uid, vals['bom_id'])
                    vals.update({'routing_id': bom_id.routing_id.id})
                else:
                    del vals['bom_id']
                res = super(mrp_production, self).write(cr, uid, ids, vals, context)
        else:
                res = super(mrp_production, self).write(cr, uid, ids, vals, context)
        return res
    
    def create(self, cr, uid, vals, context={}):
        
        if 'bom_id' in vals:
            if not 'product_def_loc' in vals:
                prod = self.pool.get('product.product').browse(cr, uid, vals['product_id'])
                vals.update({'product_def_loc': prod.default_location.id})
            if vals['bom_id']:
                bom_id = self.pool.get('mrp.bom').browse(cr, uid, vals['bom_id'])
                vals.update({'routing_id': bom_id.routing_id.id})
            else:
                del vals['bom_id']
            res = super(mrp_production, self).create(cr, uid, vals, context)
        else:
            res = super(mrp_production, self).create(cr, uid, vals, context)
        return res
    
    def onchange_bom(self, cr, uid, ids, bom_id, context=None):
        val = {}
        if bom_id:
            bom = self.pool.get('mrp.bom').browse(cr, uid, bom_id)
            val = {
                'routing_id': bom.routing_id.id
            }
        return {'value': val}
    
    def product_id_change_inh(self, cr, uid, ids, product_id, context=None):
        value = super(mrp_production, self).product_id_change(cr, uid, ids, product_id, context=None)
        val = value['value']
        if product_id:
            prod = self.pool.get('product.product').browse(cr, uid, product_id)
            val['product_def_loc'] = prod.default_location.id
        return {'value': val}
    
    def do_full_move_consume(self, cr, uid, ids, context=None):
      objs = self.pool.get('mrp.production').browse(cr , uid, ids)
      con_prod = []
      for obj in objs:
          for prod in obj.move_lines:
              con_prod.append(prod.id) 
      for prod_id in con_prod:
            if context is None:
                context = {}
            move_obj = self.pool.get('stock.move')
            line = self.pool.get('stock.move').browse(cr , uid, prod_id)
            prod_qty = line.product_qty
            prod_loc = line.location_id
            move_obj.action_consume(cr, uid, [prod_id] ,prod_qty,prod_loc.id,context=context)       
      return True
  
    def do_full_produce(self, cr, uid, ids, context=None):
      objs = self.pool.get('mrp.production').browse(cr , uid, ids)
      con_prod = []
      for obj in objs:
          for prod in obj.move_created_ids:
              con_prod.append(prod.id) 
      for prod_id in con_prod:
            if context is None:
                context = {}
            move_obj = self.pool.get('stock.move')
            line = self.pool.get('stock.move').browse(cr , uid, prod_id)
            prod_qty = line.product_qty
            prod_loc = line.location_id
            move_obj.action_consume(cr, uid, [prod_id] ,prod_qty,prod_loc.id,context=context)
      return True
  

  
    def action_compute(self, cr, uid, ids, properties=[]):
        """ Computes bills of material of a product.
        @param properties: List containing dictionaries of properties.
        @return: No. of products.
        """
        results = []
        bom_obj = self.pool.get('mrp.bom')
        prod_line_obj = self.pool.get('mrp.production.product.line')
        operation_line_obj = self.pool.get('mrp.production.operation')
        user_obj = self.pool.get('mrp.production.users')
        final_product_obj = self.pool.get('mrp.production.final.product')
        for production in self.browse(cr, uid, ids):
            prod_line_obj.unlink(cr, uid, prod_line_obj.search(cr, uid, [('production_id', '=', production.id)]))
            operation_line_obj.unlink(cr, uid, operation_line_obj.search(cr, uid, [('production_id', '=', production.id)]))
            final_product_obj.unlink(cr, uid, final_product_obj.search(cr, uid, [('production_id', '=', production.id)]))
            
            bom_point = production.bom_id
            bom_id = production.bom_id.id
            if not bom_point:
                bom_id = bom_obj._bom_find(cr, uid, production.product_id.id, production.product_uom.id, properties)
                if bom_id:
                    bom_point = bom_obj.browse(cr, uid, bom_id)
                    routing_id = bom_point.routing_id.id or False
                    self.write(cr, uid, [production.id], {'bom_id': bom_id, 'routing_id': routing_id})

            if not bom_id:
                raise osv.except_osv(_('Error'), _("Couldn't find bill of material for product"))

            factor = production.product_qty * production.product_uom.factor / bom_point.product_uom.factor
       
            res = bom_obj._bom_explode(cr, uid, bom_point, factor / bom_point.product_qty, properties)
            
            fix_obj = self.pool.get('mrp.fixed.cost')
            
            for del_it in production.fixed_costs:
                context = {'borrar':True}
                fix_obj.unlink(cr, uid, del_it.id, context=context)
            
            for aux in bom_point.fixed_costs:
                values = {
                          'name':aux.name,
                          'amount':aux.amount,
                          'production_id':production.id,
                          'estimated':aux.amount
                          }
                fix_obj.create(cr, uid, values)
            
            for user in user_obj.search(cr, uid, [('prod', '=', production.id)]):
                user_obj.unlink(cr, uid, user)
            for line in res:
                line['production_id'] = production.id
                line_products = line['products']
                del line['products']
                line_finals = line['finals']
                del line['finals']
                operation_id = operation_line_obj.create(cr, uid, line)
                for component in line_products:
                    component['mrp_production_operation_id'] = operation_id
                    component['production_id'] = production.id
                    prod_line_obj.create(cr, uid, component)
                for final_prod in line_finals:
                    final_prod['mrp_production_operation_id'] = operation_id
                    final_prod['production_id'] = production.id
                    final_product_obj.create(cr, uid, final_prod)
                self.create_operator(cr, uid, operation_line_obj.browse(cr, uid, operation_id))
        return len(results)
    
    def action_confirm(self, cr, uid, ids):
        """ Confirms production order.
        @return: Newly generated picking Id.
        """
        res_final_ids = []
        picking_id = False
        proc_ids = []
        move_obj = self.pool.get('stock.move')
        proc_obj = self.pool.get('procurement.order')
        wf_service = netsvc.LocalService("workflow")
        for production in self.browse(cr, uid, ids):
            if not production.product_lines:
                self.action_compute(cr, uid, [production.id])
                production = self.browse(cr, uid, [production.id])[0]
            pick_type = 'internal'
            moves = []
            for i in range(len(production.production_operation_ids)):
                move_id = False
                next_operation = False
                newdate = production.date_planned
                operation = production.production_operation_ids[i]
                if i < len(production.production_operation_ids)-1:
                    next_operation = production.production_operation_ids[i+1]
                for line in operation.final_products_ids:
                    if not line.product_id.default_location:
                        raise osv.except_osv(_('Product Default Location'), _('All final products must have a default location !'))
                for line in operation.components_used_ids:
                    if not line.product_id.default_location:
                        raise osv.except_osv(_('Product Default Location'), _('All components must have a default location !'))
                if not production.product_id.default_location:
                    raise osv.except_osv(_('Product Default Location'), _('All final products must have a default location !'))
                if (next_operation) and (next_operation.location_id.usage == 'supplier') and (not next_operation.location_id.address_id):
                        raise osv.except_osv(_('External Operation'), _('External operation location must be set !'))
                if (next_operation) and (next_operation.location_id.usage == 'supplier'):
                    picking_id = self.create_picking(cr, uid, 'out', 'direct', production, next_operation)
                    self.create_purchase(cr, uid, next_operation)
                    for line in next_operation.components_used_ids:
                        move_obj.create(cr, uid, {
                            'name':'PROD:' + production.name + '/'+ operation.name,
                            'date': production.date_planned,
                            'product_id': line.product_id.id,
                            'picking_id': picking_id,
                            'product_qty': line.product_qty,
                            'bom_prop': line.product_qty / production.product_qty,
                            'product_uom': line.product_uom.id,
                            'product_uos_qty': False,
                            'product_uos': False,
                            'location_id': line.product_id.default_location.id,
                            'location_dest_id': next_operation.location_id.id,
                            'state': 'confirmed',
                            'company_id': production.company_id.id,
                        })
                        proc_id = proc_obj.create(cr, uid, {
                            'name': (production.origin or '').split(':')[0] + ':' + production.name + '/' + next_operation.name,
                            'origin': (production.origin or '').split(':')[0] + ':' + production.name + '/' + next_operation.name,
                            'date_planned': newdate,
                            'product_id': line.product_id.id,
                            'product_qty': line.product_qty,
                            'product_uom': line.product_uom.id,
                            'product_uos_qty': line.product_uos and production.product_qty or False,
                            'product_uos': line.product_uos and production.product_uos.id or False,
                            'location_id': line.product_id.default_location.id,
                            'procure_method': line.product_id.procure_method,
#                            'move_id': move_id,
                            'company_id': production.company_id.id,
                        })
                        wf_service.trg_validate(uid, 'procurement.order', proc_id, 'button_confirm', cr)
                        proc_ids.append(proc_id)
                    move_pick_id = move_obj.create(cr, uid, {
                        'name':'PROD:' + production.name + '/'+ next_operation.name,
                        'date': production.date_planned,
                        'picking_id': picking_id,
                        'product_id': operation.product_id.id,
                        'product_qty': operation.semi_prod_qty,
                        'product_uom': operation.product_id.product_tmpl_id.uom_id.id,
                        'product_uos_qty': False,
                        'product_uos': False,
                        'date': newdate,
                        'location_id': operation.location_id.id,
                        'location_dest_id': next_operation.location_id.id,
                        'state': 'waiting',
                        'company_id': production.company_id.id,
                    })
                    data = move_obj.create(cr, uid, {
                            'name':'PROD:' + production.name + '/'+ operation.name,
                            'date': production.date_planned,
                            'product_id': operation.product_id.id,
                            'product_qty': operation.semi_prod_qty,
                            'product_uom': operation.product_id.product_tmpl_id.uom_id.id,
                            'product_uos_qty': False,
                            'product_uos': False,
                            'date': newdate,
                            'location_id': operation.product_id.product_tmpl_id.property_stock_production.id,
                            'location_dest_id': operation.location_id.id,
                            'move_dest_id': move_pick_id,
                            'state': 'waiting',
                            'company_id': production.company_id.id,
                            'operation_id': operation.id,
                    })
                    res_final_ids.append(data)
                    wf_service.trg_validate(uid, 'stock.picking', picking_id, 'button_confirm', cr)
                elif (next_operation) and (next_operation.location_id.usage == 'internal'):
                    if not operation.product_id:
                        raise osv.except_osv(_('Semi-elaborated Product'), _('A semi-elaborated product is missing in the %s operation !') %(operation.name))
                    semi_product_id = move_obj.create(cr, uid, {
                        'name':'PROD:' + production.name + '/'+ operation.name,
                        'product_id': operation.product_id.id,
                        'product_qty': operation.semi_prod_qty,
                        'bom_prop': operation.semi_prod_qty / production.product_qty,
                        'product_uom': operation.product_id.product_tmpl_id.uom_id.id,
                        'product_uos_qty': False,
                        'product_uos': False,
                        'date': newdate,
                        'location_id': next_operation.location_id.id or False,
                        'location_dest_id': operation.product_id.product_tmpl_id.property_stock_production.id,
                        'state': 'waiting',
                        'company_id': production.company_id.id,
                        'operation_id': next_operation.id or False,
                    })
                    moves.append(semi_product_id)
                    if not operation.location_id.usage == 'supplier':
                        data = move_obj.create(cr, uid, {
                            'name':'PROD:' + production.name + '/'+ operation.name,
                            'date': production.date_planned,
                            'product_id': operation.product_id.id,
                            'product_qty': operation.semi_prod_qty,
                            'product_uom': operation.product_id.product_tmpl_id.uom_id.id,
                            'product_uos_qty': False,
                            'product_uos': False,
                            'date': newdate,
                            'location_id': operation.product_id.product_tmpl_id.property_stock_production.id,
                            'location_dest_id': next_operation.location_id.id,
                            'move_dest_id': semi_product_id,
                            'state': 'waiting',
                            'company_id': production.company_id.id,
                            'operation_id': operation.id,
                        })
                        res_final_ids.append(data)
                if operation.location_id.usage == 'supplier':
                    picking_id = self.create_picking(cr, uid, 'in', 'direct', production, operation)
                    move_pick_id = move_obj.create(cr, uid, {
                        'name':'PROD:' + production.name + '/'+ operation.name,
                        'date': production.date_planned,
                        'picking_id': picking_id,
                        'product_id': operation.product_id.id,
                        'product_qty': operation.semi_prod_qty,
                        'product_uom': operation.product_id.product_tmpl_id.uom_id.id,
                        'product_uos_qty': False,
                        'product_uos': False,
                        'date': newdate,
                        'location_id': operation.location_id.id,
                        'location_dest_id': next_operation.location_id.id,
                        'move_dest_id': semi_product_id,
                        'state': 'draft',
                        'company_id': production.company_id.id,
                    })
                    wf_service.trg_validate(uid, 'stock.picking', picking_id, 'button_confirm', cr)
                else:  
                    if operation.components_used_ids:
                        picking_id = self.create_picking(cr, uid, pick_type, 'direct', production, operation, 'input')
                    for line in operation.components_used_ids:
                        if line.product_id.type in ('product', 'consu'):
                            move_id = move_obj.create(cr, uid, {
                                'name':'PROD:' + production.name + '/'+ operation.name,
                                'date': production.date_planned,
                                'product_id': line.product_id.id,
                                'product_qty': line.product_qty,
                                'bom_prop': line.product_qty / production.product_qty,
                                'product_uom': line.product_uom.id,
                                'product_uos_qty': line.product_uos and line.product_uos_qty or False,
                                'product_uos': line.product_uos and line.product_uos.id or False,
                                'location_id': operation.location_id.id,
                                'location_dest_id': line.product_id.product_tmpl_id.property_stock_production.id,
                                'state': 'waiting',
                                'company_id': production.company_id.id,
                                'operation_id': operation.id,
                            })
                            moves.append(move_id)
                            move_pick_id = move_obj.create(cr, uid, {
                                'name':'PROD:' + production.name + '/'+ operation.name,
                                'date': production.date_planned,
                                'product_id': line.product_id.id,
                                'picking_id': picking_id,
                                'product_qty': line.product_qty,
                                'bom_prop': line.product_qty / production.product_qty,
                                'product_uom': line.product_uom.id,
                                'product_uos_qty': line.product_uos and line.product_uos_qty or False,
                                'product_uos': line.product_uos and line.product_uos.id or False,
                                'location_id': line.product_id.default_location.id,
                                'location_dest_id': operation.location_id.id,
                                'move_dest_id': move_id,
                                'state': 'confirmed',
                                'company_id': production.company_id.id,
                            })
                        wf_service.trg_validate(uid, 'stock.picking', picking_id, 'button_confirm', cr)
                        ##TODO: Create procurements for each consumed product  
                        proc_id = proc_obj.create(cr, uid, {
                            'name': (production.origin or '').split(':')[0] + ':' + production.name + '/' + operation.name,
                            'origin': (production.origin or '').split(':')[0] + ':' + production.name + '/' + operation.name,
                            'date_planned': newdate,
                            'product_id': line.product_id.id,
                            'product_qty': line.product_qty,
                            'product_uom': line.product_uom.id,
                            'product_uos_qty': line.product_uos and production.product_qty or False,
                            'product_uos': line.product_uos and production.product_uos.id or False,
                            'location_id': line.product_id.default_location.id,
                            'procure_method': line.product_id.procure_method,
#                            'move_id': move_id,
                            'company_id': production.company_id.id,
                        })
                        wf_service.trg_validate(uid, 'procurement.order', proc_id, 'button_confirm', cr)
                        proc_ids.append(proc_id)
                    if operation.final_products_ids:
                        picking_id = self.create_picking(cr, uid, pick_type, 'direct', production, operation, 'output')
                    for line in operation.final_products_ids:
                        move_pick_id = move_obj.create(cr, uid, {
                                'name':'PROD:' + production.name + '/'+ operation.name,
                                'date': production.date_planned,
                                'product_id': line.product_id.id,
                                'picking_id': picking_id,
                                'product_qty': line.product_qty,
                                'bom_prop': line.product_qty / production.product_qty,
                                'product_uom': line.product_uom.id,
                                'product_uos_qty': False,
                                'product_uos': False,
                                'location_id': operation.location_id.id,
                                'location_dest_id': line.product_id.default_location.id,
                                'state': 'confirmed',
                                'company_id': production.company_id.id,
                            })
                        data = move_obj.create(cr, uid, {
                            'name':'PROD:' + production.name + '/'+ operation.name,
                            'date': production.date_planned,
                            'product_id': line.product_id.id,
                            'product_qty': line.product_qty,
                            'bom_prop': line.product_qty / production.product_qty,
                            'product_uom': line.product_uom.id,
                            'product_uos_qty': False,
                            'product_uos': False,
                            'location_id': line.product_id.product_tmpl_id.property_stock_production.id,
                            'location_dest_id': operation.location_id.id,
#                            'move_dest_id': move_pick_id,
                            'state': 'waiting',
                            'company_id': production.company_id.id,
                            'operation_id': operation.id,
                        })
                        res_final_ids.append(data)
                        wf_service.trg_validate(uid, 'stock.picking', picking_id, 'button_confirm', cr)
                    if not next_operation:
                        if not operation.final_products_ids:
                            picking_id = self.create_picking(cr, uid, pick_type, 'direct', production, operation, 'output')
                        move_pick_id = move_obj.create(cr, uid, {
                                'name':'PROD:' + production.name + '/'+ operation.name,
                                'date': production.date_planned,
                                'picking_id': picking_id,
                                'product_id': production.product_id.id,
                                'product_qty': production.product_qty,
                                'product_uom': production.product_uom.id,
                                'product_uos_qty': production.product_uos and production.product_uos_qty or False,
                                'product_uos': production.product_uos and production.product_uos.id or False,
                                'location_id': operation.location_id.id,
                                'location_dest_id': production.product_id.default_location.id,
                                'state': 'confirmed',
                                'company_id': production.company_id.id,
                            })
                        wf_service.trg_validate(uid, 'stock.picking', picking_id, 'button_confirm', cr)
                        product_bom = move_obj.create(cr, uid, {
                            'name':'PROD:' + production.name + '/'+ operation.name,
                            'date': production.date_planned,
                            'product_id': production.product_id.id,
                            'product_qty': production.product_qty,
                            'product_uom': production.product_uom.id,
                            'product_uos_qty': production.product_uos and production.product_uos_qty or False,
                            'product_uos': production.product_uos and production.product_uos.id or False,
                            'location_id': production.product_id.product_tmpl_id.property_stock_production.id,
                            'location_dest_id': operation.location_id.id,
#                            'move_dest_id': move_pick_id,
                            'state': 'waiting',
                            'company_id': production.company_id.id,
                            'operation_id': operation.id,
                        }) 
                        res_final_ids.append(product_bom)
                self.write(cr, uid, [production.id], {'move_created_ids': [(6, 0, res_final_ids)]})
                 
            self.write(cr, uid, [production.id], {'picking_id': picking_id, 'move_lines': [(6, 0, moves)], 'state':'confirmed'})
            message = _("Manufacturing order '%s' is scheduled for the %s.") % (
                production.name,
                datetime.strptime(production.date_planned,'%Y-%m-%d %H:%M:%S').strftime('%m/%d/%Y'),
            )
            self.log(cr, uid, production.id, message)
                
        return True
    
    def create_picking(self, cr, uid, pick_type, move_type, production, operation, internal_type=False):
        pick_obj = self.pool.get('stock.picking')
        seq_obj = self.pool.get('ir.sequence')
        pick_name = seq_obj.get(cr, uid, 'stock.picking.' + pick_type)
        picking_id = pick_obj.create(cr, uid, {
            'name': pick_name,
            'origin': (production.origin or '').split(':')[0] + ':' + production.name + '/' + operation.name,
            'type': pick_type,
            'move_type': move_type,
            'address_id': operation.location_id.address_id.id,
            'internal_type': internal_type,
            'state': 'confirmed',
            'company_id': production.company_id.id,
        })
        return picking_id
    
    def create_purchase(self, cr, uid, next_operation):
        po_obj = self.pool.get('purchase.order')
        po_lines_obj =self.pool.get('purchase.order.line')
        notes = ''
        
        notes += _('Products delivered:') + '\n\n'
        for prod in next_operation.components_used_ids:
            notes += prod.name + ' ' + str(prod.product_qty) + ' ' + prod.product_uom.name + '\n'
        notes += next_operation.product_id.name + ' ' + str(next_operation.semi_prod_qty) + ' ' + next_operation.semi_prod_uom.name
        
        po_id = po_obj.create(cr, uid, {
            'origin': next_operation.production_id.name + '/' + next_operation.name,
            'partner_id': next_operation.location_id.address_id.partner_id.id,
            'partner_address_id': next_operation.location_id.address_id.partner_id.address[0].id,
            'location_id': next_operation.location_id.id,
            'pricelist_id': next_operation.location_id.address_id.partner_id.property_product_pricelist_purchase.id,
            'notes': notes,
            'state': 'draft',
        })
        
        
        if not next_operation.workcenter_id.product_id:
            raise osv.except_osv(_('User error'), _('There is no product assigned for the external workcenter !'))
        
        seller_delay = 0
        if next_operation.workcenter_id.product_id.seller_ids:
            for s in next_operation.workcenter_id.product_id.seller_ids:
                if s.name.id == next_operation.location_id.address_id.partner_id:
                    seller_delay = s.delay
        
        po_lines_obj.create(cr, uid, {
            'name': next_operation.workcenter_id.product_id.name,
            'product_id': next_operation.workcenter_id.product_id.id,
            'product_uom': next_operation.workcenter_id.product_id.uom_id.id,
            'product_qty': next_operation.semi_prod_qty,
            'price_unit': next_operation.workcenter_id.product_id.standard_price,
            'date_planned': (datetime.now() + relativedelta(days=int(seller_delay) or 0.0)).strftime('%Y-%m-%d %H:%M:%S'),
            'order_id': po_id,
        })
        return True
    
    def create_operator(self, cr, uid, operation):
        user_obj = self.pool.get('mrp.production.users')
        for user in operation.workcenter_id.user_ids:
            user_obj.create(cr, uid, {
                'name': user.id,
                'op': operation.id,
                'prod': operation.production_id.id,
                'percent': 100,
            })
        return True
mrp_production()
